;;; COPYRIGHT 2006 Sergio Garcia <sergio.garcia@gmail.com>
;;; 
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU Lesser General Public License as published by
;;; the Free Software Foundation; either version 2.1 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA



(in-package #:gordon)

(defun draw-shape-with-style (contours
			      &key 
			      (fill-gradient-matrix nil)
			      (fill-gradient nil)
			      (fill-bitmap nil)
			      (fill-bitmap-style *clipped-bitmap-fill*)
			      (fill-bitmap-matrix (make-matrix))
			      (fill-color *a-white*)
			      (stroke-color *a-black*)
			      (stroke-width 20))

  "Creates a flash shape from a list of contours. A contour is a list of straight lines and curved lines, defined as a list of steps. An step is either an (xdelta ydelta) delta or an (xcontol ycontrol xdelta ydelta). The first step is always a (x y) and signals the start point of the contour. Draw-shape will not verify that the contours are closed."
  
  (make-shape-with-style 
   :fillstyles (make-fillstyle-array 
		:styles (cond ((and fill-gradient-matrix fill-gradient)
			       (list
				(make-fillstyle
				 :type *radial-gradient-fill*
				 :gradient-matrix fill-gradient-matrix
				 :gradient fill-gradient) 
				))
			      (fill-bitmap
			       (list (make-fillstyle :type
						     fill-bitmap-style
						     :bitmap-id fill-bitmap
						     :bitmap-matrix fill-bitmap-matrix
						     )))
			      
			      (t (list (make-fillstyle :color fill-color)))))
   

   :linestyles (make-linestyle-array 
		:styles `(,(make-linestyle 
			    :width stroke-width 
			    :color stroke-color)))

   :records (apply #'append (mapcar #'make-records-from-contour contours))))




   

(defun curved-edge-p (x)
  (= 4 (length x)))

(defun straight-edge-p (x)
  (= 2 (length x)))

(defun make-records-from-contour (contour &key (linestyle 1) (fillstyle0 1) (fillstyle1 0) (bitmap-id nil)) 
  (labels ((make-record (x)
	     (when (or 
		    (not (or (straight-edge-p x) (curved-edge-p x)))
		    (not (find-if (lambda (it) (not (zerop it))) x)))
	       (format t "Corrupt step in contour?: ~A" x))
	     (if (straight-edge-p x)
		 (make-straight-edge-record 
		  :delta-x (first x) 
		  :delta-y (second x))
		 (make-curved-edge-record  
		  :control-delta-x (first x) 
		  :control-delta-y (second x) 
		  :anchor-delta-x (third x) 
		  :anchor-delta-y (fourth x)))))
    
    (cons (make-style-change-record :delta-x (first (first contour)) 
				    :delta-y (second (first contour)) 
				    :linestyle linestyle
				    :fillstyle0 fillstyle0 
				    :fillstyle1 fillstyle1)

	  (mapcar #'make-record (cdr contour)))))

(defun make-polygone-contour (n l)
  (cons '(0 0)
	(do ((i 0 (1+ i))
	     (angle 0 (+ angle (/ (* 2 pi) n)))
	     (sides `() (append sides (list (list (round (* l (sin angle))) (round (* l (cos angle))))))))
	    ((= i n) sides))))

(defun make-arch (n r a &key (initial-angle 0))
  "segments, radio, angle"
  (let* ((angle (/ a n))
	 (l (* r (tan (/ angle 2)))))
    (do ((i 0 (1+ i))
	       (total-angle initial-angle (+ total-angle angle))
	       (sides `() (append sides (list (list 
					       (round (* l (cos total-angle))) 
					       (round (* l (sin total-angle))) 
					       (round (* l (cos (+ angle total-angle)))) 
					       (round (* l (sin (+ angle total-angle)))))))))
	      ((= i n) sides))))

(defun make-circle-contour (n r)
  "Creates a circular contour of radio r, using n steps"
  (cons '(0 0) (make-arch n r (* 2 pi))))



(defun make-bridge-contour (steps radio angle width &key (initial-angle 0)) 
  (close-contour 
  (append 
   `((,(round (* radio (sin initial-angle))) ,(- (round (* radio (cos initial-angle))))))
   (make-arch steps radio angle :initial-angle initial-angle)
   `((,(round (* width (cos (+ initial-angle angle (/ pi 2))))) ,(round (* width (sin (+ initial-angle angle (/ pi 2)))))))
   (make-arch steps (- radio width) (* -1 angle) :initial-angle (- (+ initial-angle angle) (* 2 pi)))
   `((,(round (* width (cos (- initial-angle (/ pi 2))))) ,(round (* width (sin (- initial-angle (/ pi 2))))))))))


(defun closed-p (contour)
  (let ((delta
	 (reduce (lambda (total step) 
		   (if (straight-edge-p step) 
		       (cons (- (car total) (first step))
			     (- (cdr total) (second step)))
		       (cons (- (car total) (first step) (third step))
			     (- (cdr total) (second step) (fourth step))))) 
		 (cdr contour) :initial-value '(0 . 0))))
    (= 0  (car delta) (cdr delta)) ))

(defun add-joint-to-contour (contour) 
  "given a contour, returns a contour that is closed, adding an straight edge if necessary"
  (let ((delta
	 (reduce (lambda (total step) 
		   (if (straight-edge-p step) 
		       (cons (- (car total) (first step))
			     (- (cdr total) (second step)))
		       (cons (- (car total) (first step) (third step))
			     (- (cdr total) (second step) (fourth step))))) 
		 (cdr contour) :initial-value '(0 . 0))))
    (if (not (= 0  (car delta) (cdr delta)))
	(append contour (list (list (car delta) (cdr delta))))
	contour)))


(defun close-contour (contour) 
  "given a contour, returns a contour that is closed, by changing the last edge"
  (let ((delta
	 (reduce (lambda (total step) 
		   (if (straight-edge-p step) 
		       (cons (- (car total) (first step))
			     (- (cdr total) (second step)))
		       (cons (- (car total) (first step) (third step))
			     (- (cdr total) (second step) (fourth step))))) 
		 (cdr contour) :initial-value '(0 . 0)))
	(last-edge (car (last contour))))
	
    (if (not (= 0  (car delta) (cdr delta)))
	(append (subseq contour 0 (- (length contour) 1))
		(if (curved-edge-p last-edge)
		    (list (list (first last-edge) (second last-edge)
				(+ (car delta) (third last-edge))
				(+ (cdr delta) (fourth last-edge))))
		    (list (list (+ (first last-edge) (car delta))
				(+ (second last-edge) (cdr delta))))))
	contour)))


(defun invert-contour-vertical (contour)
  (mapcar (lambda (step) (mapcar-pos (lambda (item pos) (if (oddp pos) (* -1 item) item)) step)) contour))

(defun scale-contour (contour xscale yscale)
  (let ((new-contour 
	 (mapcar (lambda (step) (if (straight-edge-p step)
				    (list (floor (* xscale (first step)))
					  (floor (* yscale (second step))))
				    (list (floor (* xscale (first step)))
					  (floor (* yscale (second step)))
					  (floor (* xscale (third step)))
					  (floor (* yscale (fourth step))))))
		 contour)))
    (if (closed-p contour)
	(close-contour new-contour)
	new-contour)))



(defun scale-contours (contours xscale yscale)
  (mapcar (lambda (contour) (scale-contour contour xscale yscale)) contours))

(defun final-point (contour)
  (reduce (lambda (current step) (if (curved-edge-p step) 
				     (cons (+ (car current) (first step) (third step))
					   (+ (cdr current) (second step) (fourth step)))
				     (cons (+ (car current) (first step))
					   (+ (cdr current) (second step))))) 
	  contour 
	  :initial-value '(0 . 0)))



