;;; COPYRIGHT 2006 Sergio Garcia <sergio.garcia@gmail.com>
;;; 
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU Lesser General Public License as published by
;;; the Free Software Foundation; either version 2.1 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA



(in-package #:gordon)

(defparameter *tag-data-marshallers* (make-hash-table))
(defparameter *tag-codes* (make-hash-table))

;;; TAGS marshalling framework
;;; we use a hashtable of lambdas instead of defmethod
;;; just to see.

(defun tag-type (tg)
  (gethash (type-of tg) *tag-codes*))

(defun marshall-tag-data (tg)
  (funcall (gethash (type-of tg) *tag-data-marshallers*) tg))

(defmacro add-tag-type(name code fields marshaller)
  `(progn (defstruct ,name ,@fields)
	 (setf (gethash ',name *tag-codes*) ,code)
	 (setf (gethash ',name *tag-data-marshallers*) ,marshaller)
	 ))

;;; Marshall a tag
(defun marshall-tag (tg)
  (let* ((mtd (flat-marshall-data (marshall-tag-data tg)))
	 (tt (tag-type tg))
	 (l (length mtd))
	 (long? (>= l 63))
	 (header-size (if long? 6 2))
	 (recordheader (if long? 
			   (list 
			    (is2leubitaba `(,tt 63) '(10 6))
			    (i2ui32 l))
			   (is2leubitaba `(,tt ,l) '(10 6)))))
    (xappend recordheader mtd)))

(defun marshall-movie-object (mo)
  (cond ((tag-type mo)
	 (marshall-tag mo))
	((streamp mo) (let ((seq (make-array (file-length mo) :element-type '(unsigned-byte 8)))) (read-sequence seq mo) seq))
	))


;;;
;;; Define all the different tag types
;;;


(add-tag-type tag-end
	      0
	      ()
	      (lambda (x) *empty-ba*))

(add-tag-type tag-showframe
	      1
	      ()
	      (lambda (x) *empty-ba*))


(add-tag-type tag-SBC 
	      9 
	      (bgcolor)
	      (lambda (x)
		(marshall-rgb (tag-SBC-bgcolor x))))


(add-tag-type tag-framelabel
	      43
	      (name)
	      (lambda (x) (string2string (tag-framelabel-name x) t)))


(add-tag-type tag-define-shape
	      32
	      (id shape)
	      (lambda (x) (xappend
			   (i2ui16 (tag-define-shape-id x))
			   (marshall-rect (calculate-shape-bounds (tag-define-shape-shape x)))
			   (marshall-shape-with-style (tag-define-shape-shape x)))))

(add-tag-type tag-place-object2
	      26
	      ((depth nil) (id nil) (move nil) (matrix nil) (cx nil) (name nil))
	      (lambda (x)
		(xappend 
		 *no-bit*
		 *no-bit*
		 (bif (tag-place-object2-name x))
		 *no-bit*
		 (bif (tag-place-object2-cx x))
		 (bif (tag-place-object2-matrix x))
		 (bif (tag-place-object2-id x))
		 (bif (tag-place-object2-move x))
		 (i2ui16 (tag-place-object2-depth x))
		 (when (tag-place-object2-id x) 
		   (i2ui16 (tag-place-object2-id x)))
		 (when (tag-place-object2-matrix x)
		   (marshall-matrix (tag-place-object2-matrix x)))
		 (when (tag-place-object2-cx x)
		   (marshall-cxform-wa (tag-place-object2-cx x)))
		 (when (tag-place-object2-name x)
		   (string2string (tag-place-object2-name x) t)))))

(add-tag-type tag-remove-object
	      28
	      ((depth nil))
	      (lambda (x) 
		(i2ui16 (tag-remove-object-depth x))))

(add-tag-type tag-define-font-2
	      48
	      ((id 0 :type integer)
	       (name "stupid-font" :type string)
	       (offsets nil) ;defined in defineFont
	       (shapes nil) ;defined in defineFont
	       (codes nil)
	       (ascent 1000 :type integer)
	       (descent 100 :type integer)
	       (leading 100 :type integer)
	       (advance 800)
	       )

	      (lambda (font)
		(let ((marshalled-shapes
		       (mapcar (lambda (x) (flat-marshall-data (marshall-shape x))) (tag-define-font-2-shapes font))) 
		      (numglyphs (length (tag-define-font-2-shapes font)))
		      )
		  (xappend
		   (i2ui16 (tag-define-font-2-id font))
		   *yes-bit* ;has layout
		   *no-bit* ;JIS
		   *no-bit* ;small text
		   *no-bit* ;ANSI
		   *yes-bit* ;wide offsets
		   *yes-bit* ;wide codes
		   *no-bit* ;italic
		   *no-bit* ;bold
		   (i2ui8 1) ; langage code latin
		   (i2ui8 (array-dimension (tag-define-font-2-name font) 0))
		   (string2string (tag-define-font-2-name font))
		   (i2ui16 numglyphs) ;numglyphs
		   (let ((total 0))
		     (loop
		      for i = 0 then (1+ i)
		      for marshalled-shape in marshalled-shapes
		      collecting (+ 
				  (* 4 (1+ numglyphs))
				  total) into offsets
		      summing (length marshalled-shape) into total
		      finally (return (mapcar #'i2ui32 offsets))
		      )) ;offsets
		   (i2ui32 (+ (* 4 (1+ numglyphs)) (reduce (lambda (total item) (+ total (length (flat-marshall-data item)))) marshalled-shapes :initial-value 0))) ;code table offset
		   (apply #'xappend marshalled-shapes)
		   (mapcar #'i2ui16 (tag-define-font-2-codes font)) ; the codes
		   
		   (i2si16 (tag-define-font-2-ascent font))
		   (i2si16 (tag-define-font-2-descent font))
		   (i2si16 (tag-define-font-2-leading font))
		   (mapcar #'i2si16 (tag-define-font-2-advance font))
		   ; (make-list numglyphs :initial-element (i2si16 (tag-define-font-2-advance font)))
		   (apply #'xappend (make-list numglyphs :initial-element (xappend (marshall-rect (make-rect :xmin 0 :xmax 1000 :ymin 0 :ymax 100)) *empty-ba*) ))
		   ;(apply #'xappend (make-list numglyphs :initial-element (marshall-rect (make-rect :xmin 0 :xmax 100 :ymin 0 :ymax 100))  ))
		   (i2ui16 0) 
		   ))))


(add-tag-type tag-define-edit-text 
	      37
	      ((id nil)
	       (word-wrap t)
	       (bounds (make-rect :xmin 0 :xmax 2000 :ymin 0 :ymax 2000))
	       (password nil)
	       (editable nil)
	       (selectable nil)
	       (border nil)
	       (html nil)
	       (use-outlines t)
	       (font-id nil)
	       (font-height 12)
	       (color *a-black*)
	       (autosize nil)
	       (max-length nil)
	       (multi-lines nil)
	       (align nil)
	       (left-margin 0)
	       (right-margin 0)
	       (indent 0)
	       (leading 0)
	       (variable-name nil)
	       (initial-text "Gordon Rules"))
	      (lambda (texto)
		(xappend
		 (i2ui16 (tag-define-edit-text-id texto))
		 (marshall-rect (tag-define-edit-text-bounds texto))
		 *empty-ba* 
					; the former is necesary to force byte-alignment
					; when are structures byte aligned? I guess they always 
					; are, except if they are part
					; of an array. Im not yet sure this is true, tho.
		 
		 
		 (bif (tag-define-edit-text-initial-text texto))
		 (bif (tag-define-edit-text-word-wrap texto))
		 (bif (tag-define-edit-text-multi-lines texto))
		 (bif (tag-define-edit-text-password texto))
		 (bif (not (tag-define-edit-text-editable texto)))
		 (bif (tag-define-edit-text-color texto))
		 (bif (tag-define-edit-text-max-length texto))
		 (bif (tag-define-edit-text-font-id texto))
		 *no-bit* ; reserved
		 (bif (tag-define-edit-text-autosize texto))
		 (bif (or 
		       (tag-define-edit-text-align texto)
		       (not (zerop (tag-define-edit-text-left-margin texto)))
		       (not (zerop (tag-define-edit-text-right-margin texto)))
		       (not (zerop (tag-define-edit-text-indent texto)))
		       (not (zerop (tag-define-edit-text-leading texto))))) ; layout
		 (bif (not (tag-define-edit-text-selectable texto)))
		 (bif (tag-define-edit-text-border texto))
		 *no-bit* ;reserved
		 (bif (tag-define-edit-text-html texto))
		 (bif (tag-define-edit-text-use-outlines texto))
		 (with id (tag-define-edit-text-font-id texto) 
		   (xappend 
		    (i2ui16 id)
		    (i2ui16 (tag-define-edit-text-font-height texto))))
		 (with color (tag-define-edit-text-color texto) (marshall-rgba color))
		 (with max (tag-define-edit-text-max-length texto) (i2ui16 max))
		 (when (or 
		       (tag-define-edit-text-align texto)
		       (not (zerop (tag-define-edit-text-left-margin texto)))
		       (not (zerop (tag-define-edit-text-right-margin texto)))
		       (not (zerop (tag-define-edit-text-indent texto)))
		       (not (zerop (tag-define-edit-text-leading texto))))
		   (xappend 
		    (i2ui8 (case (tag-define-edit-text-align texto)
			     (:LEFT 0)
			     (:RIGHT 1)
			     (:CENTER 2)
			     (:JUSTIFY 3)
			     (nil 2)))
		    (i2ui16 (tag-define-edit-text-left-margin texto))
		    (i2ui16 (tag-define-edit-text-right-margin texto))
		    (i2ui16 (tag-define-edit-text-indent texto))
		    (i2ui16 (tag-define-edit-text-leading texto))))
		 (string2string (tag-define-edit-text-variable-name texto) t)
		 (with init-text (tag-define-edit-text-initial-text texto)
		       
		   (string2string init-text t)))))


(add-tag-type tag-do-action
	      12
	      (records)
	      (lambda (x) (xappend
				 (apply #'xappend (mapcar #'marshall-action-record (tag-do-action-records x)))
				 (i2ui8 0))
			   ))

(add-tag-type tag-define-bits
	      6
	      ((id nil)
	       (jpegdata nil)
	       )
	      (lambda (x)
		(xappend
		 (i2ui16 (tag-define-bits-id x))
		 (tag-define-bits-jpegdata x))
		))


(add-tag-type tag-define-bits2
	      21
	      ((id nil)
	       (jpegdata nil)
	       )
	      (lambda (x)
		(xappend
		 (i2ui16 (tag-define-bits2-id x))
		 (tag-define-bits2-jpegdata x))
		))



(add-tag-type tag-define-button-2
	      34
	      ((id nil)
	       (menu nil)
	       (action-offset nil)
	       (characters nil)
	       (actions nil)
	       )
	      (lambda (tdb)
		(let* ((characters (apply #'xappend (mapcar #'marshall-button-record
							   (tag-define-button-2-characters tdb)
							   )))
		      (marshalled-characters (flat-marshall-data characters))
		      )
		  
		  (xappend
		   (i2ui16 (tag-define-button-2-id tdb))
		   (i2ubita 0 7)
		   (bif (tag-define-button-2-menu tdb))
		   ;
		   (if (zerop (length (tag-define-button-2-actions tdb)))
		       (i2ui16 0) (i2ui16 (length marshalled-characters))
		       )
 
		   ;(i2ui16 0)
		   
		   marshalled-characters
		   (i2ui8 0)))))


;(marshall-tag (make-tag-do-action :records (list (make-action-record-next-frame))))

		 
(defun make-jpeg-tag (id file)
  (with-open-file (stream file :direction :input :element-type '(unsigned-byte 8))
    (let ((seq (make-array (file-length stream) :element-type '(unsigned-byte 8)))) (read-sequence seq stream)
	 (make-tag-define-bits2 :id id :jpegdata seq)
	 )))





