;;; COPYRIGHT 2006 Sergio Garcia <sergio.garcia@gmail.com>
;;; 
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU Lesser General Public License as published by
;;; the Free Software Foundation; either version 2.1 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA



(in-package #:gordon)

;;;; Structures of "higher" SWF types

(defstruct rect
  (xmin 0 :type integer)
  (xmax 0 :type integer)
  (ymin 0 :type integer)
  (ymax 0 :type integer))

(defstruct rgb
  (r 0 :type integer) (g 0 :type integer) (b 0 :type integer))
  
(defstruct rgba
  (r 0 :type integer) (g 0 :type integer) (b 0 :type integer) (a 0 :type integer))

;; color transformation with alpha
(defstruct cxform-wa 
  (rmult 1.0 :type float)
  (gmult 1.0 :type float)
  (bmult 1.0 :type float)
  (amult 1.0 :type float)
  (radd 0 :type integer)
  (gadd 0 :type integer)
  (badd 0 :type integer)
  (aadd 0 :type integer))

(defun cxwa-has-mult (cx)
  (not (and (= (cxform-wa-rmult cx) 1)
	   (= (cxform-wa-gmult cx) 1)
	   (= (cxform-wa-bmult cx) 1)
	   (= (cxform-wa-amult cx) 1))))

(defun cxwa-has-add (cx)
  (not (and (zerop (cxform-wa-radd cx))
	   (zerop (cxform-wa-gadd cx))
	   (zerop (cxform-wa-badd cx))
	   (zerop (cxform-wa-aadd cx)))))


(defun marshall-cxform-wa (cx)
  (let ((hm (cxwa-has-mult cx))
	(hd (cxwa-has-add cx)))
    (xappend 
     (bif hd)
     (bif hm)
     (when (or hm hd)
       (is2besbitawithmax 
	(append
	 (when hm (mapcar (lambda (n) (floor (* 256 n))) 
			  `(,(cxform-wa-rmult cx)
			    ,(cxform-wa-gmult cx)
			    ,(cxform-wa-bmult cx)
			    ,(cxform-wa-amult cx))))
	 (when hd `(,(cxform-wa-radd cx)
		    ,(cxform-wa-gadd cx)
		    ,(cxform-wa-badd cx)
		    ,(cxform-wa-aadd cx))))
	4)))))

(deftype color () `(or rgb rgba))

(defstruct matrix 
  (scale-x 1.0 :type float)
  (scale-y 1.0 :type float)
  (rotate-skew-0 0.0 :type float)
  (rotate-skew-1 0.0 :type float)
  (translate-x 0 :type integer)
  (translate-y 0 :type integer))

(defstruct gradient-record
  (ratio 1 :type integer)
  (color (make-rgb) :type color))

(defstruct gradient
  (record-list '() :type list))

; Shapes 

(defstruct fillstyle
  (type *solid-fill* :type integer)
  (color nil :type (or color null))
  (gradient-matrix nil :type (or matrix null))
  (gradient nil :type (or gradient null))
  (bitmap-id nil :type (or integer null))
  (bitmap-matrix nil :type (or matrix null)))

(defstruct linestyle
  (width 1 :type integer)
  (color nil :type (or color null)))

(defstruct fillstyle-array
  (styles '() :type list))

(defstruct linestyle-array
  (styles '() :type list))

(defstruct style-change-record 
  (delta-x 0 :type integer)
  (delta-y 0 :type integer)
  (fillstyle0 nil :type (or null integer))
  (fillstyle1 nil :type (or null integer))
  (linestyle nil :type (or null integer)))

(defstruct straight-edge-record
  (delta-x 0 :type integer)
  (delta-y 0 :type integer))
  
(defstruct curved-edge-record
  (control-delta-x 0 :type integer)
  (control-delta-y 0 :type integer)
  (anchor-delta-x 0 :type integer)
  (anchor-delta-y 0 :type integer))
	  
(defstruct end-record)

(deftype shape-record () 
  `(or 
    style-change-record
    straight-edge-record
    curved-edge-record
    end-record))

(defstruct shape
    (records nil :type (or list null)))

(defstruct shape-with-style
  (fillstyles nil :type (or fillstyle-array null))
  (linestyles nil :type (or linestyle-array null))
  (records nil :type (or list null)))


(defstruct button-record
  (test-hit nil)
  (state-down nil)
  (state-over nil)
  (state-up nil)
  (character-id nil)
  (depth nil)
  (place-matrix nil :type matrix)
  (color-transform nil :type cxform-wa)
  )

(defun calculate-shape-bounds (shape)
  "Calculate the minimum rectangle (with sides parallel to the axis) that
contains a shape. This is used by the DEFINE SHAPE TAG. If no rectangle is defined, the refresh may fail."

  (let ((bounds (make-rect))
	(x 0)
	(y 0))
    (dolist (record (shape-with-style-records shape) bounds)
      (typecase record 
	(style-change-record 
	 (setf x (+ x (style-change-record-delta-x record)))
	 (setf y (+ y (style-change-record-delta-y record)))
	 (setf (rect-xmin bounds) (min x (rect-xmin bounds)))
	 (setf (rect-xmax bounds) (max x (rect-xmax bounds)))
	 (setf (rect-ymin bounds) (min y (rect-ymin bounds)))
	 (setf (rect-ymax bounds) (max y (rect-ymax bounds))))
	
	(straight-edge-record 
	 (setf x (+ x (straight-edge-record-delta-x record)))
	 (setf y (+ y (straight-edge-record-delta-y record)))
	 (setf (rect-xmin bounds) (min x (rect-xmin bounds)))
	 (setf (rect-xmax bounds) (max x (rect-xmax bounds)))
	 (setf (rect-ymin bounds) (min y (rect-ymin bounds)))
	 (setf (rect-ymax bounds) (max y (rect-ymax bounds))))
		
	(curved-edge-record
	 (let ((ix (+ x (curved-edge-record-control-delta-x record)))
	       (iy (+ y (curved-edge-record-control-delta-y record))))
	   (setf x (+ ix (curved-edge-record-anchor-delta-x record)))
	   (setf y (+ iy (curved-edge-record-anchor-delta-y record)))
	   (setf (rect-xmin bounds) (min x ix (rect-xmin bounds)))
	   (setf (rect-xmax bounds) (max x ix (rect-xmax bounds)))
	   (setf (rect-ymin bounds) (min y iy (rect-ymin bounds)))
	   (setf (rect-ymax bounds) (max y iy (rect-ymax bounds)))))))))


;;;; Multiply transformation matrix
(defun make-rotation-matrix-angle (angle)
  (make-matrix :scale-x (cos angle)
	       :scale-y (cos angle)
	       :rotate-skew-0 (* -1 (sin angle))
	       :rotate-skew-1 (sin angle)))


(defun +-matrix (matrix1 matrix2)
  (make-matrix
   :scale-x (+ (matrix-scale-x matrix1) (matrix-scale-x matrix2))
   :scale-y (+ (matrix-scale-y matrix1) (matrix-scale-y matrix2))
   :rotate-skew-0 (+ (matrix-rotate-skew-0 matrix1) (matrix-rotate-skew-0 matrix2))
   :rotate-skew-1 (+ (matrix-rotate-skew-1 matrix1) (matrix-rotate-skew-1 matrix2))
   :translate-x (+ (matrix-translate-x matrix1) (matrix-translate-x matrix2))
   :translate-y (+ (matrix-translate-y matrix1) (matrix-translate-y matrix2))))

(defun *-matrix (matrix1 matrix2)
  (make-matrix
   :scale-x (+ (* (matrix-scale-x matrix1) (matrix-scale-x matrix2))
	       (* (matrix-rotate-skew-0 matrix1) (matrix-rotate-skew-1 matrix2)))
   :scale-y (+ (* (matrix-rotate-skew-1 matrix1) (matrix-rotate-skew-0 matrix2))
	       (* (matrix-scale-y matrix1) (matrix-scale-y matrix2)))
   :rotate-skew-0 (+ (* (matrix-scale-x matrix1) (matrix-rotate-skew-0 matrix2))
		     (* (matrix-rotate-skew-0 matrix1) (matrix-scale-y matrix2)))
   :rotate-skew-1 (+ (* (matrix-rotate-skew-1 matrix1) (matrix-scale-x matrix2))
		     (* (matrix-scale-x matrix1) (matrix-scale-y matrix2)))
   :translate-x (floor (+ (* (matrix-scale-x matrix1) (matrix-translate-x matrix2))
			   (* (matrix-rotate-skew-0 matrix1) (matrix-translate-y matrix2))
			   (matrix-translate-x matrix1)))
   :translate-y (floor (+ (* (matrix-scale-y matrix1) (matrix-translate-y matrix2))
			   (* (matrix-rotate-skew-1 matrix1) (matrix-translate-x matrix2))
			   (matrix-translate-y matrix1)))))



;;;;
;;;; Marshalling of SWF objects
;;;; the type returned for all marshall functions is
;;;; a list that contains byte arrays and bit arrays
;;;; 

;;; this functions takes a list that represents marshalled data, and 
;;; returns the corresponsing SWF byte array, byte-aligned

(defun flat-marshall-data (md)
  (apply #'concba (nreverse 
		   (reduce #'push-val (xappend md (list *empty-ba*)) :initial-value '())
		   )))

;(defun push-val (lst val) 
;  (cond ((typep (car lst) '(or null byte-array)) (push val lst))
;	((and (typep (car lst) 'bit-array) (typep val 'bit-array))
;	 (push-val (cdr lst) (concbita (car lst) val)))
;	((and (typep (car lst) 'bit-array) (typep val 'byte-array))
;	 (push-val (cdr lst) (concba (bita2beba (car lst)) val)))
;	(t (error (format nil "unrecognized type ~A ~A " (type-of (car lst)) (type-of val)) ))))

(defun push-val (lst val) 
  (cond ((typep (car lst) '(or null byte-array)) (push val lst))
	((and (typep (car lst) 'bit-array) (typep val 'bit-array))
	 (let ((ato (concbita (pop lst) val)))
	   (push-val lst ato)))
	((and (typep (car lst) 'bit-array) (typep val 'byte-array))
	 (let ((ato (bita2beba (pop lst))))
	   (push-val lst (concba ato val))))
	(t (error (format nil "unrecognized type ~A ~A " (type-of (car lst)) (type-of val)) ))))




(defun marshall-rect (rect)
  (xappend (is2besbitawithmax  `(,(rect-xmin rect)
			,(rect-xmax rect)
			,(rect-ymin rect)
			,(rect-ymax rect))
		      5)
	   ))



(defun marshall-color (col)
  (typecase col
    (rgb (marshall-rgb col))
    (rgba (marshall-rgba col))))
  

(defun marshall-rgb (col)
  (xappend (i2ui8 (rgb-r col))
	  (i2ui8 (rgb-g col))
	  (i2ui8 (rgb-b col))))

(defun marshall-rgba (col)
  (xappend (i2ui8 (rgba-r col))
	   (i2ui8 (rgba-g col))
	   (i2ui8 (rgba-b col))
	   (i2ui8 (rgba-a col))))

(defun marshall-gradient-record (gr)
  (xappend (i2ui8 (gradient-record-ratio gr))
	  (marshall-color (gradient-record-color gr))
	  ))

(defun marshall-gradient (g)
  (apply #'xappend (i2ui8 (length (gradient-record-list g)))
	 `(,@(mapcar #'marshall-gradient-record (gradient-record-list g)))))


(defun marshall-button-record (br)
  (xappend
   (i2ubita 0 4)
   (bif (button-record-test-hit br))
   (bif (button-record-state-down br))
   (bif (button-record-state-over br))
   (bif (button-record-state-up br))
   (i2ui16 (button-record-character-id br))
   (i2ui16 (button-record-depth br))
   (marshall-matrix (button-record-place-matrix br))
   *empty-ba*
   (marshall-cxform-wa (button-record-color-transform br))
   *empty-ba*
   )
  )

(defun marshall-fillstyle (fs)
  (xappend
   (i2ui8 (fillstyle-type fs))
   (if (= (fillstyle-type fs) *solid-fill*)
       (marshall-color (fillstyle-color fs))
       *empty-ba*)
   (if (or (= (fillstyle-type fs) *linear-gradient-fill*)
	   (= (fillstyle-type fs) *radial-gradient-fill*))
       (xappend (marshall-matrix (fillstyle-gradient-matrix fs))
		(marshall-gradient (fillstyle-gradient fs)))
       (xappend *empty-bita* *empty-ba*))
   (if (find (fillstyle-type fs) `(,*repeating-bitmap-fill* 
				   ,*clipped-bitmap-fill*
				   ,*non-smoothed-repeating-bitmap*
				   ,*non-smoothed-clipped-bitmap*))
       (xappend (i2ui16 (fillstyle-bitmap-id fs))
		(marshall-matrix (fillstyle-bitmap-matrix fs))
		*empty-ba*
		)
       (xappend  *empty-bita* *empty-ba*))))


(defun marshall-fillstyle-array (fa)
  (let ((styles (fillstyle-array-styles fa)))
    (xappend 
     (if (< (length styles) 255)
	 (i2ui8 (length styles))
	 (xappend (i2ui8 255) (i2ui16 (length styles))))
     (apply #'xappend (mapcar #'marshall-fillstyle styles)))))

(defun marshall-linestyle (ls)
  (xappend (i2ui16 (linestyle-width ls))
	   (marshall-color (linestyle-color ls))))

(defun marshall-linestyle-array (la)
  (let ((styles (linestyle-array-styles la)))
    (xappend 
     (if (< (length styles) 255)
	 (i2ui8 (length styles))
	 (xappend (i2ui8 255) (i2ui16 (length styles))))
     (apply #'xappend (mapcar #'marshall-linestyle styles)))))



;;; marshall a matrix
(defun matrix-has-scale (m)
  (not (and (= (matrix-scale-x m) 1)
	    (= (matrix-scale-y m) 1))))

(defun matrix-has-rotation (m)
  (not (and (zerop (matrix-rotate-skew-1 m))
	    (zerop (matrix-rotate-skew-0 m)))))

(defun matrix-has-translation (m)
  (not (and (zerop (matrix-translate-x m))
	    (zerop (matrix-translate-y m)))))


(defun marshall-matrix (m)
  (xappend
   (if (matrix-has-scale m)
       (xappend *yes-bit*
		 (fs2besbitawithmax 
		  `(,(matrix-scale-x m)
		    ,(matrix-scale-y m)) 5))
       *no-bit*)
   (if (matrix-has-rotation m)
       (xappend *yes-bit*
		 (fs2besbitawithmax 
		  `(,(matrix-rotate-skew-0 m)
		    ,(matrix-rotate-skew-1 m)) 5))
       *no-bit*)
   ;;(if (matrix-has-translation m)
   ;;    (xappend *yes-bit*
   (is2besbitawithmax 
    `(,(matrix-translate-x m)
      ,(matrix-translate-y m)) 5))
)
    ;;   *no-bit*)))


;;; shapes marshalling

(defun style-change-record-has-delta (sc)
  (or
   (not (zerop (style-change-record-delta-x sc)))
   (not (zerop (style-change-record-delta-y sc)))))

(defun marshall-style-change-record (sr fs-count ls-count)
  (xappend 
   *no-bit* ;no edge
   *no-bit* ;no new styles
   (bif (style-change-record-linestyle sr))
   (bif (style-change-record-fillstyle1 sr))
   (bif (style-change-record-fillstyle0 sr))
   (if (style-change-record-has-delta sr)
     (xappend *yes-bit*
	      (is2besbitawithmax `(,(style-change-record-delta-x sr)
				   ,(style-change-record-delta-y sr))
				 5))
     *no-bit*)
   
   (when (style-change-record-fillstyle0 sr)
     (i2ubita (style-change-record-fillstyle0 sr) fs-count))

   (when (style-change-record-fillstyle1 sr)
     (i2ubita (style-change-record-fillstyle1 sr) fs-count))
   
   (when (style-change-record-linestyle sr) 
     (i2ubita (style-change-record-linestyle sr) ls-count))))

(defun marshall-end-record ()
    (xappend
     *no-bit*
     (i2ubita 0 5)))


(defun marshall-curved-edge-record (er)
    (let ((neededsize  (max 
	     (needsize (curved-edge-record-control-delta-x er) t)
	     (needsize (curved-edge-record-control-delta-y er) t)
	     (needsize (curved-edge-record-anchor-delta-x er) t)
	     (needsize (curved-edge-record-anchor-delta-y er) t))))
      (xappend 
       *yes-bit* ;edge
       *no-bit*  ;not straight
       (i2ubita (- neededsize 2) 4)
       (i2sbita (curved-edge-record-control-delta-x er) neededsize)
       (i2sbita (curved-edge-record-control-delta-y er) neededsize)
       (i2sbita (curved-edge-record-anchor-delta-x er) neededsize)
       (i2sbita (curved-edge-record-anchor-delta-y er) neededsize))))

(defun marshall-straight-edge-record (er)
  (let ((neededsize  (max 
		      2
		      (needsize (straight-edge-record-delta-x er) t)
		      (needsize (straight-edge-record-delta-y er) t))))
    
    (xappend 
     *yes-bit* ;edge record
     *yes-bit* ;straight
     (i2ubita (- neededsize 2) 4)
     (cond ((not (or (zerop (straight-edge-record-delta-x er))
	       (zerop (straight-edge-record-delta-y er))))
	    (xappend 
	     *yes-bit*
	     (i2sbita (straight-edge-record-delta-x er) neededsize)
	     (i2sbita (straight-edge-record-delta-y er) neededsize)))
	   ((zerop (straight-edge-record-delta-x er))
	    (xappend
	     *no-bit*
	     *yes-bit*
	     (i2sbita (straight-edge-record-delta-y er) neededsize)))
	   (t 
	    (xappend
	     *no-bit*
	     *no-bit*
	     (i2sbita (straight-edge-record-delta-x er) neededsize)))))))
   

(defun marshall-shape-record(sr fs-count ls-count) 
  (typecase sr 
    (style-change-record (marshall-style-change-record sr fs-count ls-count))
    (straight-edge-record (marshall-straight-edge-record sr))
    (curved-edge-record (marshall-curved-edge-record sr))
    (end-record (marshall-end-record))))


(defun marshall-shape (shp)
  (let* ((fillbits (needsize 1 nil))
	 (linebits (needsize 1 nil)))
    (xappend 
     (i2ubita fillbits 4)
     (i2ubita linebits 4)
     (apply #'xappend (mapcar (lambda (x) (marshall-shape-record x fillbits linebits)) (append (shape-records shp) (list (make-end-record)))))
     )))

(defun marshall-shape-with-style (sws)
  (let* ((fs-count (length (fillstyle-array-styles (shape-with-style-fillstyles sws))))
	 (ls-count (length (linestyle-array-styles (shape-with-style-linestyles sws))))
	 (fillbits (needsize fs-count nil))
	 (linebits (needsize ls-count nil)))
	  
    (xappend 
     (marshall-fillstyle-array (shape-with-style-fillstyles sws))
     (marshall-linestyle-array (shape-with-style-linestyles sws))
     (i2ubita fillbits 4)
     (i2ubita linebits 4)
     (apply #'xappend (mapcar (lambda (x) (marshall-shape-record x fillbits linebits)) (append (shape-with-style-records sws) (list (make-end-record)))))
     )))

